# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from anyio import Lock, connect_tcp


class Websocket:
    def __init__(self, websocket, path: str):
        self._websocket = websocket
        self._path = path
        self._send_lock = Lock()

    @property
    def path(self) -> str:
        return self._path

    def __aiter__(self):
        return self

    async def __anext__(self) -> bytes:
        try:
            message = await self.recv()
        except Exception:
            raise StopAsyncIteration()
        return message

    async def send(self, message: bytes):
        async with self._send_lock:
            await self._websocket.send_bytes(message)

    async def recv(self) -> bytes:
        b = await self._websocket.receive_bytes()
        return bytes(b)


async def ensure_server_running(host: str, port: int) -> None:
    while True:
        try:
            await connect_tcp(host, port)
        except OSError:
            pass
        else:
            break



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

import json
import subprocess
from functools import partial
from pathlib import Path

import pytest
from anyio import Event, create_task_group
from hypercorn import Config
from hypercorn.asyncio import serve
from pycrdt_websocket import ASGIServer, WebsocketServer
from utils import ensure_server_running

# workaround until these PRs are merged:
# - https://github.com/yjs/y-websocket/pull/104


def update_json_file(path: Path, d: dict):
    with open(path, "rb") as f:
        package_json = json.load(f)
    package_json.update(d)
    with open(path, "w") as f:
        json.dump(package_json, f, indent=2)


here = Path(__file__).parent
d = {"type": "module"}
update_json_file(here.parent / "node_modules/y-websocket/package.json", d)


@pytest.fixture
async def yws_server(request, unused_tcp_port):
    try:
        async with create_task_group() as tg:
            try:
                kwargs = request.param
            except Exception:
                kwargs = {}
            websocket_server = WebsocketServer(**kwargs)
            app = ASGIServer(websocket_server)
            config = Config()
            config.bind = [f"localhost:{unused_tcp_port}"]
            shutdown_event = Event()
            async with websocket_server as websocket_server:
                tg.start_soon(
                    partial(serve, app, config, shutdown_trigger=shutdown_event.wait, mode="asgi")
                )
                await ensure_server_running("localhost", unused_tcp_port)
                pytest.port = unused_tcp_port
                yield unused_tcp_port, websocket_server
                shutdown_event.set()
    except Exception:
        pass


@pytest.fixture
def yjs_client(request):
    client_id = request.param
    p = subprocess.Popen(["node", f"{here / 'yjs_client_'}{client_id}.js", str(pytest.port)])
    yield p
    p.terminate()
    try:
        p.wait(timeout=10)
    except Exception:
        p.kill()



# Copyright (c) Jupyter Development Team.
# Distributed under the terms of the Modified BSD License.

from pycrdt import Awareness, Doc

from jupyter_ydoc import YBlob, YNotebook


def test_yblob():
    yblob = YBlob()
    assert yblob.get() == b""
    yblob.set(b"012")
    assert yblob.get() == b"012"
    changes = []

    def callback(topic, event):
        changes.append((topic, event))

    yblob.observe(callback)
    yblob.set(b"345")
    assert len(changes) == 1
    topic, event = changes[0]
    assert topic == "source"
    assert event.keys["bytes"]["oldValue"] == b"012"
    assert event.keys["bytes"]["newValue"] == b"345"


def test_ynotebook_undo_manager():
    ynotebook = YNotebook()
    cell0 = {
        "cell_type": "code",
        "source": "Hello",
    }
    ynotebook.append_cell(cell0)
    source = ynotebook.ycells[0]["source"]
    source += ", World!\n"
    cell1 = {
        "cell_type": "code",
        "source": "print(1 + 1)\n",
    }
    ynotebook.append_cell(cell1)
    assert len(ynotebook.ycells) == 2
    assert str(ynotebook.ycells[0]["source"]) == "Hello, World!\n"
    assert str(ynotebook.ycells[1]["source"]) == "print(1 + 1)\n"
    assert ynotebook.undo_manager.can_undo()
    ynotebook.undo_manager.undo()
    assert len(ynotebook.ycells) == 1
    assert str(ynotebook.ycells[0]["source"]) == "Hello, World!\n"
    assert ynotebook.undo_manager.can_undo()
    ynotebook.undo_manager.undo()
    assert len(ynotebook.ycells) == 1
    assert str(ynotebook.ycells[0]["source"]) == "Hello"
    assert ynotebook.undo_manager.can_undo()
    ynotebook.undo_manager.undo()
    assert len(ynotebook.ycells) == 0
    assert not ynotebook.undo_manager.can_undo()


def test_awareness():
    yblob = YBlob()
    assert yblob.awareness is None

    ydoc = Doc()
    awareness = Awareness(ydoc)
    yblob = YBlob(ydoc, awareness)
    assert yblob.awareness == awareness
